from typing import Any, Self

from pydantic import BaseModel, model_validator

from ..utils import cleanup_text, remove_ruby_tags, replace_device_params


class APIModel(BaseModel):
    """Base class for all models in hakushin-py."""

    @property
    def fields(self) -> dict[str, Any]:
        """Return all fields of the model as a dictionary."""
        schema = self.model_fields
        field_names = schema.keys()
        field_values = {name: getattr(self, name) for name in field_names}
        return field_values

    @model_validator(mode="after")
    def _format_fields(self) -> Self:
        for field_name, field_value in self.fields.items():
            if field_name in {"name", "description", "story"} and isinstance(field_value, str):
                setattr(
                    self,
                    field_name,
                    replace_device_params(remove_ruby_tags(cleanup_text(field_value))),
                )

        return self



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ..constants import HSR_API_LANG_MAP, TRAILBLAZER_NAMES
from ..enums import Game, Language
from ..models import hsr
from ..utils import cleanup_text, remove_ruby_tags, replace_placeholders
from .base import BaseClient

if TYPE_CHECKING:
    from aiohttp import ClientSession

__all__ = ("HSRClient",)


class HSRClient(BaseClient):
    """Client to interact with the Hakushin Honkai Star Rail API."""

    def __init__(
        self,
        lang: Language = Language.EN,
        *,
        cache_path: str = "./.cache/hakushin/aiohttp-cache.db",
        cache_ttl: int = 3600,
        headers: dict[str, Any] | None = None,
        debug: bool = False,
        session: ClientSession | None = None,
    ) -> None:
        super().__init__(
            Game.HSR,
            lang,
            cache_path=cache_path,
            cache_ttl=cache_ttl,
            headers=headers,
            debug=debug,
            session=session,
        )

    async def fetch_new(self, *, use_cache: bool = True) -> hsr.New:
        """Fetch the ID of beta items in Honkai Star Rail.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A model representing the new items.
        """
        data = await self._request("new", use_cache, static=True)
        return hsr.New(**data)

    async def fetch_characters(self, *, use_cache: bool = True) -> list[hsr.Character]:
        """Fetch all Honkai Star Rail characters.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of character objects.
        """
        data = await self._request("character", use_cache, in_data=True)

        characters = [hsr.Character(id=int(char_id), **char) for char_id, char in data.items()]
        for char in characters:
            char.name = remove_ruby_tags(char.names[HSR_API_LANG_MAP[self.lang]])
            if char.name == "{NICKNAME}":
                char.name = TRAILBLAZER_NAMES[self.lang]

        return characters

    async def fetch_character_detail(
        self, character_id: int, *, use_cache: bool = True
    ) -> hsr.CharacterDetail:
        """Fetch the details of a Honkai Star Rail character.

        Args:
            character_id: The character ID.
            use_cache: Whether to use the response cache.

        Returns:
            The character details object.
        """
        endpoint = f"character/{character_id}"
        data = await self._request(endpoint, use_cache)
        return hsr.CharacterDetail(**data)

    async def fetch_light_cones(self, *, use_cache: bool = True) -> list[hsr.LightCone]:
        """Fetch all Honkai Star Rail light cones.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of light cone objects.
        """
        endpoint = "lightcone"
        data = await self._request(endpoint, use_cache, in_data=True)
        light_cones = [
            hsr.LightCone(id=int(light_cone_id), **light_cone)
            for light_cone_id, light_cone in data.items()
        ]
        for light_cone in light_cones:
            light_cone.name = remove_ruby_tags(light_cone.names[HSR_API_LANG_MAP[self.lang]])
        return light_cones

    async def fetch_light_cone_detail(
        self, light_cone_id: int, *, use_cache: bool = True
    ) -> hsr.LightConeDetail:
        """Fetch the details of a Honkai Star Rail light cone.

        Args:
            light_cone_id: The light cone ID.
            use_cache: Whether to use the response cache.

        Returns:
            The light cone details object.
        """
        endpoint = f"lightcone/{light_cone_id}"
        data = await self._request(endpoint, use_cache)
        return hsr.LightConeDetail(**data)

    async def fetch_relic_sets(self, *, use_cache: bool = True) -> list[hsr.RelicSet]:
        """Fetch all Honkai Star Rail relic sets.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of relic set objects.
        """
        endpoint = "relicset"
        data = await self._request(endpoint, use_cache, in_data=True)
        sets = [hsr.RelicSet(id=int(set_id), **set_) for set_id, set_ in data.items()]

        for set_ in sets:
            set_.name = remove_ruby_tags(set_.names[HSR_API_LANG_MAP[self.lang]])
            two_piece = set_.set_effect.two_piece
            two_piece.description = replace_placeholders(
                cleanup_text(two_piece.descriptions[HSR_API_LANG_MAP[self.lang]]),
                two_piece.parameters,
            )
            if (four_piece := set_.set_effect.four_piece) is not None:
                four_piece.description = replace_placeholders(
                    cleanup_text(four_piece.descriptions[HSR_API_LANG_MAP[self.lang]]),
                    four_piece.parameters,
                )

        return sets

    async def fetch_relic_set_detail(
        self, set_id: int, *, use_cache: bool = True
    ) -> hsr.RelicSetDetail:
        """Fetch the details of a Honkai Star Rail relic set.

        Args:
            set_id: The relic set ID.
            use_cache: Whether to use the response cache.

        Returns:
            The relic set details object.
        """
        endpoint = f"relicset/{set_id}"
        data = await self._request(endpoint, use_cache)
        return hsr.RelicSetDetail(**data)



from enum import IntEnum, StrEnum

__all__ = ("Game", "Language")


class Game(StrEnum):
    """Games supported by the Hakushin API."""

    GI = "gi"
    """Genshin Impact."""
    HSR = "hsr"
    """Honkai: Star Rail."""
    ZZZ = "zzz"
    """Zenless Zone Zero."""


class Language(StrEnum):
    """Lanauges supported by the Hakushin API."""

    EN = "en"
    """English."""
    ZH = "zh"
    """Simple Chinese."""
    KO = "ko"
    """Korean."""
    JA = "ja"
    """Japanese."""


class GIElement(StrEnum):
    """Genshin Impact element."""

    HYDRO = "Hydro"
    PYRO = "Pyro"
    CRYO = "Cryo"
    ELECTRO = "Electro"
    ANEMO = "Anemo"
    GEO = "Geo"
    DENDRO = "Dendro"


class HSRElement(StrEnum):
    """HSR element."""

    ICE = "Ice"
    FIRE = "Fire"
    THUNDER = "Thunder"
    WIND = "Wind"
    PHYSICAL = "Physical"
    QUANTUM = "Quantum"
    IMAGINARY = "Imaginary"


class HSRPath(StrEnum):
    """HSR character path."""

    PRESERVATION = "Knight"
    THE_HUNT = "Rogue"
    DESTRUCTION = "Warrior"
    ERUDITION = "Mage"
    HARMONY = "Shaman"
    NIHILITY = "Warlock"
    ABUNDANCE = "Priest"


class ZZZSpecialty(IntEnum):
    """ZZZ character specialty."""

    ATTACK = 1
    STUN = 2
    ANOMALY = 3
    SUPPORT = 4
    DEFENSE = 5


class ZZZElement(IntEnum):
    """ZZZ character element."""

    PHYSICAL = 200
    FIRE = 201
    ICE = 202
    ELECTRIC = 203
    ETHER = 205


class ZZZAttackType(IntEnum):
    """ZZZ character attack type."""

    SLASH = 101
    STRIKE = 102
    PIERCE = 103


class ZZZSkillType(StrEnum):
    """ZZZ character skill type."""

    BASIC = "Basic"
    DODGE = "Dodge"
    SPECIAL = "Special"
    CHAIN = "Chain"
    ASSIST = "Assist"



from typing import Any, Literal

from pydantic import Field, computed_field, field_validator, model_validator

from ...constants import ZZZ_SA_RARITY_CONVERTER
from ...enums import ZZZAttackType, ZZZElement, ZZZSkillType, ZZZSpecialty
from ...utils import cleanup_text
from ..base import APIModel
from .common import ZZZExtraProp, ZZZMaterial

__all__ = (
    "CharaCoreSkillLevel",
    "CharaSkillDescParam",
    "CharaSkillDescParamProp",
    "Character",
    "CharacterAscension",
    "CharacterCoreSkill",
    "CharacterDetail",
    "CharacterExtraAscension",
    "CharacterInfo",
    "CharacterProp",
    "CharacterSkill",
    "CharacterSkillDesc",
    "MindscapeCinema",
)


class Character(APIModel):
    """ZZZ character (agent)."""

    id: int
    name: str = Field(alias="code")
    rarity: Literal["S", "A"] | None = Field(alias="rank")
    specialty: ZZZSpecialty = Field(alias="type")
    element: ZZZElement | None
    attack_type: ZZZAttackType | None = Field(alias="hit")
    image: str = Field(alias="icon")
    en_description: str = Field(alias="desc")
    names: dict[Literal["EN", "KO", "CHS", "JA"], str]

    @computed_field
    @property
    def phase_3_cinema_art(self) -> str:
        """Agent phase 3 mindscape cinema art.

        Example: https://api.hakush.in/zzz/UI/Mindscape_1041_3.webp
        """
        return f"https://api.hakush.in/zzz/UI/Mindscape_{self.id}_3.webp"

    @computed_field
    @property
    def phase_2_cinema_art(self) -> str:
        """Agent phase 2 mindscape cinema art.

        Example: https://api.hakush.in/zzz/UI/Mindscape_1041_2.webp
        """
        return f"https://api.hakush.in/zzz/UI/Mindscape_{self.id}_2.webp"

    @computed_field
    @property
    def phase_1_cinema_art(self) -> str:
        """Agent phase 1 mindscape cinema art.

        Example: https://api.hakush.in/zzz/UI/Mindscape_1041_1.webp
        """
        return f"https://api.hakush.in/zzz/UI/Mindscape_{self.id}_1.webp"

    @computed_field
    @property
    def icon(self) -> str:
        """Agent icon.

        Example: https://api.hakush.in/zzz/UI/IconRoleSelect01.webp
        """
        return self.image.replace("Role", "RoleSelect")

    @field_validator("rarity", mode="before")
    @classmethod
    def __convert_rarity(cls, value: int | None) -> Literal["S", "A"] | None:
        return ZZZ_SA_RARITY_CONVERTER[value] if value is not None else None

    @field_validator("attack_type", mode="before")
    @classmethod
    def __convert_attack_type(cls, value: int) -> ZZZAttackType | None:
        try:
            return ZZZAttackType(value)
        except ValueError:
            return None

    @field_validator("image")
    @classmethod
    def __convert_image(cls, value: str) -> str:
        return f"https://api.hakush.in/zzz/UI/{value}.webp"

    @model_validator(mode="before")
    @classmethod
    def __pop_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["names"] = {
            "EN": values.pop("EN"),
            "KO": values.pop("KO"),
            "CHS": values.pop("CHS"),
            "JA": values.pop("JA"),
        }
        return values


class CharacterProp(APIModel):
    """ZZZ character property."""

    id: int
    name: str

    @model_validator(mode="before")
    @classmethod
    def __transform(cls, values: dict[str, Any]) -> dict[str, Any]:
        first_item = next(iter(values.items()))
        return {"id": first_item[0], "name": first_item[1]}


class CharacterInfo(APIModel):
    """ZZZ character detail info."""

    birthday: str = Field(alias="Birthday")
    full_name: str = Field(alias="FullName")
    gender: str = Field(alias="Gender")
    female_impression: str = Field(alias="ImpressionF")
    male_impression: str = Field(alias="ImpressionM")
    outlook_desc: str = Field(alias="OutlookDesc")
    profile_desc: str = Field(alias="ProfileDesc")
    faction: str = Field(alias="Race")
    unlock_conditions: list[str] = Field(alias="UnlockCondition")

    @field_validator("female_impression", "male_impression", "outlook_desc", "profile_desc")
    @classmethod
    def __cleanup_text(cls, value: str) -> str:
        return cleanup_text(value)


class MindscapeCinema(APIModel):
    """ZZZ character mindscape cinema (constellation)."""

    level: int = Field(alias="Level")
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    description2: str = Field(alias="Desc2")

    @field_validator("description", "description2")
    @classmethod
    def __cleanup_text(cls, value: str) -> str:
        return cleanup_text(value)


class CharacterAscension(APIModel):
    """A ZZZ character ascension object."""

    max_hp: int = Field(alias="HpMax")
    attack: int = Field(alias="Attack")
    defense: int = Field(alias="Defence")
    max_level: int = Field(alias="LevelMax")
    min_level: int = Field(alias="LevelMin")
    materials: list[ZZZMaterial] = Field(alias="Materials")

    @field_validator("materials", mode="before")
    @classmethod
    def __convert_materials(cls, value: dict[str, int]) -> list[ZZZMaterial]:
        return [ZZZMaterial(id=int(k), amount=v) for k, v in value.items()]


class CharacterExtraAscension(APIModel):
    """ZZZ character extra ascension object."""

    max_level: int = Field(alias="MaxLevel")
    props: list[ZZZExtraProp] = Field(alias="Extra")

    @field_validator("props", mode="before")
    @classmethod
    def __convert_props(cls, value: dict[str, dict[str, Any]]) -> list[ZZZExtraProp]:
        return [ZZZExtraProp(**data) for data in value.values()]


class CharaSkillDescParamProp(APIModel):
    """ZZZ character skill description parameter property."""

    main: int = Field(alias="Main")
    growth: int = Field(alias="Growth")
    format: str = Field(alias="Format")


class CharaSkillDescParam(APIModel):
    """ZZZ character skill description parameter."""

    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    params: dict[str, CharaSkillDescParamProp] | None = Field(None, alias="Param")


class CharacterSkillDesc(APIModel):
    """ZZZ character skill description."""

    name: str = Field(alias="Name")
    description: str | None = Field(None, alias="Desc")
    params: list[CharaSkillDescParam] | None = Field(None, alias="Param")


class CharacterSkill(APIModel):
    """ZZZ character skill."""

    descriptions: list[CharacterSkillDesc] = Field(alias="Description")
    materials: dict[str, list[ZZZMaterial]] = Field(alias="Material")
    type: ZZZSkillType = Field(alias="Type")

    @field_validator("materials", mode="before")
    @classmethod
    def __convert_materials(cls, value: dict[str, dict[str, int]]) -> dict[str, list[ZZZMaterial]]:
        return {
            k: [ZZZMaterial(id=int(k), amount=v) for k, v in data.items()]
            for k, data in value.items()
        }


class CharaCoreSkillLevel(APIModel):
    """One level of a ZZZ character core skill."""

    level: int = Field(alias="Level")
    id: int = Field(alias="Id")
    names: list[str] = Field(alias="Name")
    descriptions: list[str] = Field(alias="Desc")

    @field_validator("descriptions")
    @classmethod
    def __cleanup_text(cls, value: list[str]) -> list[str]:
        return [cleanup_text(v) for v in value]


class CharacterCoreSkill(APIModel):
    """ZZZ character core skill."""

    levels: dict[int, CharaCoreSkillLevel] = Field(alias="Level")
    """Key is the level of the core skill."""
    level_up_materials: dict[str, list[ZZZMaterial]] | None = Field(None, alias="Materials")

    @field_validator("level_up_materials", mode="before")
    @classmethod
    def __convert_materials(cls, value: dict[str, dict[str, int]]) -> dict[str, list[ZZZMaterial]]:
        return {
            k: [ZZZMaterial(id=int(k), amount=v) for k, v in data.items()]
            for k, data in value.items()
        }

    @field_validator("levels", mode="before")
    @classmethod
    def __intify_keys(cls, value: dict[str, dict[str, Any]]) -> dict[int, CharaCoreSkillLevel]:
        return {int(k): CharaCoreSkillLevel(**v) for k, v in value.items()}


class CharacterDetail(APIModel):
    """ZZZ character detail."""

    id: int = Field(alias="Id")
    image: str = Field(alias="Icon")
    name: str = Field(alias="Name")
    code_name: str = Field(alias="CodeName")
    rarity: Literal["S", "A"] | None = Field(alias="Rarity")
    specialty: CharacterProp = Field(alias="WeaponType")
    element: CharacterProp = Field(alias="ElementType")
    attack_type: CharacterProp = Field(alias="HitType")
    faction: CharacterProp = Field(alias="Camp")
    gender: Literal["M", "F"] = Field(alias="Gender")
    info: CharacterInfo | None = Field(alias="PartnerInfo")
    stats: dict[str, float] = Field(alias="Stats")
    mindscape_cinemas: list[MindscapeCinema] = Field(alias="Talent")
    ascension: list[CharacterAscension] = Field(alias="Level")
    extra_ascension: list[CharacterExtraAscension] = Field(alias="ExtraLevel")
    skills: dict[ZZZSkillType, CharacterSkill] = Field(alias="Skill")
    passive: CharacterCoreSkill = Field(alias="Passive")

    @computed_field
    @property
    def phase_3_cinema_art(self) -> str:
        """Agent phase 3 mindscape cinema art.

        Example: https://api.hakush.in/zzz/UI/Mindscape_1041_3.webp
        """
        return f"https://api.hakush.in/zzz/UI/Mindscape_{self.id}_3.webp"

    @computed_field
    @property
    def phase_2_cinema_art(self) -> str:
        """Agent phase 2 mindscape cinema art.

        Example: https://api.hakush.in/zzz/UI/Mindscape_1041_2.webp
        """
        return f"https://api.hakush.in/zzz/UI/Mindscape_{self.id}_2.webp"

    @computed_field
    @property
    def phase_1_cinema_art(self) -> str:
        """Agent phase 1 mindscape cinema art.

        Example: https://api.hakush.in/zzz/UI/Mindscape_1041_1.webp
        """
        return f"https://api.hakush.in/zzz/UI/Mindscape_{self.id}_1.webp"

    @computed_field
    @property
    def icon(self) -> str:
        """Character icon.

        Example: https://api.hakush.in/zzz/UI/IconRoleSelect01.webp
        """
        return self.image.replace("Role", "RoleSelect")

    @field_validator("info", mode="before")
    @classmethod
    def __convert_info(cls, value: dict[str, Any]) -> CharacterInfo | None:
        return None if not value else CharacterInfo(**value)

    @field_validator("skills", mode="before")
    @classmethod
    def __convert_skills(
        cls, value: dict[str, dict[str, Any]]
    ) -> dict[ZZZSkillType, CharacterSkill]:
        return {
            ZZZSkillType(k): CharacterSkill(Type=ZZZSkillType(k), **v) for k, v in value.items()
        }

    @field_validator("extra_ascension", mode="before")
    @classmethod
    def __convert_extra_ascension(
        cls, value: dict[str, dict[str, Any]]
    ) -> list[CharacterExtraAscension]:
        return [CharacterExtraAscension(**data) for data in value.values()]

    @field_validator("ascension", mode="before")
    @classmethod
    def __convert_ascension(cls, value: dict[str, dict[str, Any]]) -> list[CharacterAscension]:
        return [CharacterAscension(**data) for data in value.values()]

    @field_validator("mindscape_cinemas", mode="before")
    @classmethod
    def __dict_to_list(cls, value: dict[str, dict[str, Any]]) -> list[MindscapeCinema]:
        return [MindscapeCinema(**data) for data in value.values()]

    @field_validator("stats", mode="before")
    @classmethod
    def __pop_tags(cls, value: dict[str, Any]) -> dict[str, float]:
        value.pop("Tags", None)
        return value

    @field_validator("gender", mode="before")
    @classmethod
    def __transform_gender(cls, value: int) -> Literal["M", "F"]:
        # Hope I don't get cancelled for this.
        # Female is '2' btw.
        return "M" if value == 1 else "F"

    @field_validator("rarity", mode="before")
    @classmethod
    def __convert_rarity(cls, value: int | None) -> Literal["S", "A"] | None:
        return ZZZ_SA_RARITY_CONVERTER[value] if value is not None else None

    @field_validator("image")
    @classmethod
    def __convert_image(cls, value: str) -> str:
        return f"https://api.hakush.in/zzz/UI/{value}.webp"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from hakushin.enums import Game, Language

from ..constants import GI_LANG_MAP
from ..models import gi
from .base import BaseClient

if TYPE_CHECKING:
    from aiohttp import ClientSession

__all__ = ("GIClient",)


class GIClient(BaseClient):
    """Client to interact with the Hakushin Genshin Impact API."""

    def __init__(
        self,
        lang: Language = Language.EN,
        *,
        cache_path: str = "./.cache/hakushin/aiohttp-cache.db",
        cache_ttl: int = 3600,
        headers: dict[str, Any] | None = None,
        debug: bool = False,
        session: ClientSession | None = None,
    ) -> None:
        super().__init__(
            Game.GI,
            lang,
            cache_path=cache_path,
            cache_ttl=cache_ttl,
            headers=headers,
            debug=debug,
            session=session,
        )

    async def fetch_new(self, *, use_cache: bool = True) -> gi.New:
        """Fetch the ID of beta items in Genshin Impact.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A model representing the new items.
        """
        data = await self._request("new", use_cache, static=True)
        return gi.New(**data)

    async def fetch_characters(self, *, use_cache: bool = True) -> list[gi.Character]:
        """Fetch all Genshin Impact characters.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of character objects.
        """
        data = await self._request("character", use_cache, in_data=True)

        characters = [gi.Character(id=char_id, **char) for char_id, char in data.items()]
        for char in characters:
            char.name = char.names[GI_LANG_MAP[self.lang]]

        return characters

    async def fetch_character_detail(
        self, character_id: str, *, use_cache: bool = True
    ) -> gi.CharacterDetail:
        """Fetch the details of a Genshin Impact character.

        Args:
            character_id: The character ID.
            use_cache: Whether to use the response cache.

        Returns:
            The character details object.
        """
        endpoint = f"character/{character_id}"
        data = await self._request(endpoint, use_cache)
        return gi.CharacterDetail(**data)

    async def fetch_weapons(self, *, use_cache: bool = True) -> list[gi.Weapon]:
        """Fetch all Genshin Impact weapons.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of weapon objects.
        """
        endpoint = "weapon"
        data = await self._request(endpoint, use_cache, in_data=True)
        weapons = [gi.Weapon(id=int(weapon_id), **weapon) for weapon_id, weapon in data.items()]
        for weapon in weapons:
            weapon.name = weapon.names[GI_LANG_MAP[self.lang]]
        return weapons

    async def fetch_weapon_detail(
        self, weapon_id: int, *, use_cache: bool = True
    ) -> gi.WeaponDetail:
        """Fetch the details of a Genshin Impact weapon.

        Args:
            weapon_id: The weapon ID.
            use_cache: Whether to use the response cache.

        Returns:
            The weapon details object.
        """
        endpoint = f"weapon/{weapon_id}"
        data = await self._request(endpoint, use_cache)
        return gi.WeaponDetail(**data)

    async def fetch_artifact_sets(self, *, use_cache: bool = True) -> list[gi.ArtifactSet]:
        """Fetch all Genshin Impact artifact sets.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of artifact set objects.
        """
        endpoint = "artifact"
        data = await self._request(endpoint, use_cache, in_data=True)
        sets = [gi.ArtifactSet(**set_) for set_ in data.values()]
        for set_ in sets:
            set_.name = set_.names[GI_LANG_MAP[self.lang]]
            set_.set_effect.two_piece.name = set_.set_effect.two_piece.names[GI_LANG_MAP[self.lang]]
            set_.set_effect.two_piece.description = set_.set_effect.two_piece.descriptions[
                GI_LANG_MAP[self.lang]
            ]
            if set_.set_effect.four_piece is not None:
                set_.set_effect.four_piece.name = set_.set_effect.four_piece.names[
                    GI_LANG_MAP[self.lang]
                ]
                set_.set_effect.four_piece.description = set_.set_effect.four_piece.descriptions[
                    GI_LANG_MAP[self.lang]
                ]
        return sets

    async def fetch_artifact_set_detail(
        self, set_id: int, *, use_cache: bool = True
    ) -> gi.ArtifactSetDetail:
        """Fetch the details of a Genshin Impact artifact set.

        Args:
            set_id: The artifact set ID.
            use_cache: Whether to use the response cache.

        Returns:
            The artifact set details object.
        """
        endpoint = f"artifact/{set_id}"
        data = await self._request(endpoint, use_cache)
        return gi.ArtifactSetDetail(**data)



__all__ = ("HakushinError", "NotFoundError")


class HakushinError(Exception):
    """Base class for exceptions in the Hakushin API."""

    def __init__(self, status: int, message: str, url: str) -> None:
        super().__init__(message)
        self.status = status
        self.message = message
        self.url = url

    def __str__(self) -> str:
        return f"{self.status}: {self.message} ({self.url})"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} status={self.status} message={self.message!r} url={self.url!r}>"


class NotFoundError(HakushinError):
    """Raised when the requested resource is not found."""

    def __init__(self, url: str) -> None:
        super().__init__(404, "The requested resource was not found.", url)



from __future__ import annotations

import re
from typing import TYPE_CHECKING, TypeVar

from .constants import (
    ASCENDED_LEVEL_TO_ASCENSION,
    ASCENSION_TO_MAX_LEVEL,
    NOT_ASCENDED_LEVEL_TO_ASCENSION,
    PERCENTAGE_FIGHT_PROPS,
    STAT_TO_FIGHT_PROP,
)
from .enums import Game

if TYPE_CHECKING:
    from .models import gi, hsr


def format_num(digits: int, calculation: float) -> str:
    """Format a number to a string with a fixed number of digits after the decimal point.

    Args:
        digits (int): Number of digits after the decimal point.
        calculation (float): Number to format.

    Returns:
        str: The formatted number.
    """
    return f"{calculation:.{digits}f}"


def replace_layout(text: str) -> str:
    """Replace the layout in a string with the corresponding word.

    Args:
        text (str): The text to format.

    Returns:
        str: The formatted text.
    """
    if "LAYOUT" in text:
        brackets = re.findall(r"{LAYOUT.*?}", text)
        word_to_replace = re.findall(r"{LAYOUT.*?#(.*?)}", brackets[0])[0]
        text = text.replace("".join(brackets), word_to_replace)
    return text


def replace_params(text: str, param_list: list[float]) -> list[str]:
    """Replace parameters in a string with the corresponding values.

    Args:
        text (str): The text to replace the parameters in.
        param_list (list[float]): The list of parameters to replace the values with.

    Returns:
        list[str]: The list of strings with the replaced parameters.
    """
    params: list[str] = re.findall(r"{[^}]*}", text)

    for item in params:
        if "param" not in item:
            continue

        param_text = re.findall(r"{param(\d+):([^}]*)}", item)[0]
        param, value = param_text

        if value in {"F1P", "F2P"}:
            result = format_num(int(value[1]), param_list[int(param) - 1] * 100)
            text = re.sub(re.escape(item), f"{result}%", text)
        elif value in {"F1", "F2"}:
            result = format_num(int(value[1]), param_list[int(param) - 1])
            text = re.sub(re.escape(item), result, text)
        elif value == "P":
            result = format_num(0, param_list[int(param) - 1] * 100)
            text = re.sub(re.escape(item), f"{result}%", text)
        elif value == "I":
            result = int(param_list[int(param) - 1])
            text = re.sub(re.escape(item), str(round(result)), text)

    text = replace_layout(text)
    text = text.replace("{NON_BREAK_SPACE}", "")
    text = text.replace("#", "")
    return text.split("|")


def replace_device_params(text: str) -> str:
    """Replace device parameters in a string with the corresponding values."""
    # Replace '{LAYOUT_CONSOLECONTROLLER#stick}' with 'stick/'
    text = re.sub(r"{LAYOUT_CONSOLECONTROLLER#(.*?)}", r"\1/", text)

    # Replace '{LAYOUT_FALLBACK#joystick}' with 'joystick'
    text = re.sub(r"{LAYOUT_FALLBACK#(.*?)}", r"\1", text)

    return text


def cleanup_text(text: str) -> str:
    """Remove HTML tags and sprite presets from a string.

    Args:
        text (str): The text to clean.

    Returns:
        str: The cleaned text.
    """
    clean = re.compile(r"<.*?>|\{SPRITE_PRESET#[^\}]+\}")
    return re.sub(clean, "", text).replace("\\n", "\n").replace("\r\n", "\n")


def replace_placeholders(text: str, param_list: list[float]) -> str:
    """Replaces placeholders in the given text with values from the parameter list.

    Args:
        text (str): The text containing placeholders to be replaced.
        param_list (list[float]): The list of parameter values.

    Returns:
        str: The text with placeholders replaced by their corresponding values.
    """
    placeholders: list[str] = re.findall(r"#\d+\[[^\]]+\]%?", text)

    for placeholder in placeholders:
        index = int(placeholder[1])
        format_ = placeholder[-1]
        value = param_list[index - 1]
        if format_ == "%":
            value *= 100
        text = text.replace(placeholder, f"{round(value)}{'%' if format_ == '%' else ''}")

    return text


def get_ascension_from_level(level: int, ascended: bool, game: Game) -> int:
    """Get the ascension from the level and ascended status."""
    if not ascended and level in NOT_ASCENDED_LEVEL_TO_ASCENSION[game]:
        return NOT_ASCENDED_LEVEL_TO_ASCENSION[game][level]

    for (start, end), ascension in ASCENDED_LEVEL_TO_ASCENSION[game].items():
        if start <= level <= end:
            return ascension

    return 0


def get_max_level_from_ascension(ascension: int, game: Game) -> int:
    """Get the max level from the ascension."""
    return ASCENSION_TO_MAX_LEVEL[game][ascension]


def calc_gi_chara_upgrade_stat_values(
    character: gi.CharacterDetail,
    level: int,
    ascended: bool,
) -> dict[str, float]:
    """Calculate the stat values of a GI character at a certain level and ascension status.

    Args:
        character: The character to calculate the stats for.
        level: The level of the character.
        ascended: Whether the character is ascended.
    """
    result: dict[str, float] = {}

    result["FIGHT_PROP_BASE_HP"] = character.base_hp * character.stats_modifier.hp[str(level)]
    result["FIGHT_PROP_BASE_ATTACK"] = character.base_atk * character.stats_modifier.atk[str(level)]
    result["FIGHT_PROP_BASE_DEFENSE"] = (
        character.base_def * character.stats_modifier.def_[str(level)]
    )

    ascension = get_ascension_from_level(level, ascended, Game.GI)
    ascension = character.stats_modifier.ascension[ascension - 1]
    for fight_prop, value in ascension.items():
        stat = STAT_TO_FIGHT_PROP.get(fight_prop, fight_prop)
        if stat not in result:
            result[stat] = 0
        result[stat] += value

    return result


def calc_hsr_chara_upgrade_stat_values(
    character: hsr.CharacterDetail,
    level: int,
    ascended: bool,
) -> dict[str, float]:
    """Calculate the stat values of a HSR character at a certain level and ascension status.

    Args:
        character: The character to calculate the stats for.
        level: The level of the character.
        ascended: Whether the character is ascended.
    """
    result: dict[str, float] = {}

    ascension = get_ascension_from_level(level, ascended, Game.HSR)
    stats = character.ascension_stats[str(ascension)]

    result["baseHP"] = stats["HPBase"] + stats["HPAdd"] * (level - 1)
    result["baseAttack"] = stats["AttackBase"] + stats["AttackAdd"] * (level - 1)
    result["baseDefence"] = stats["DefenceBase"] + stats["DefenceAdd"] * (level - 1)

    result["baseSpeed"] = stats["SpeedBase"]
    result["criticalChanceBase"] = stats["CriticalChance"]
    result["criticalDamageBase"] = stats["CriticalDamage"]

    return result


def calc_weapon_upgrade_stat_values(
    weapon: gi.WeaponDetail,
    level: int,
    ascended: bool,
) -> dict[str, float]:
    """Calculate the stat values of a GI weapon at a certain level and ascension.

    Args:
        weapon: The weapon to calculate the stats for.
        level: The level of the weapon.
        ascended: Whether the weapon is ascended.
    """
    result: dict[str, float] = {}

    result["FIGHT_PROP_BASE_ATTACK"] = (
        weapon.stat_modifiers["ATK"].base * weapon.stat_modifiers["ATK"].levels[str(level)]
    )

    for fight_prop, value in weapon.stat_modifiers.items():
        if fight_prop == "ATK":
            continue
        result[fight_prop] = value.base * value.levels[str(level)]

    ascension = get_ascension_from_level(level, ascended, Game.GI)
    if ascension == 0:
        ascension = 1
    ascension = weapon.ascension[str(ascension)]
    for fight_prop, value in ascension.items():
        if fight_prop not in result:
            result[fight_prop] = 0
        result[fight_prop] += value

    return result


def calc_light_cone_upgrade_stat_values(
    light_cone: hsr.LightConeDetail,
    level: int,
    ascended: bool,
) -> dict[str, float]:
    """Calculate the stat values of a HSR light cone at a certain level and ascension status.

    Args:
        light_cone: The light cone to calculate the stats for.
        level: The level of the light cone.
        ascended: Whether the light cone is ascended.
    """
    result: dict[str, float] = {}

    ascension = get_ascension_from_level(level, ascended, Game.HSR)
    stats = light_cone.ascension_stats[ascension]

    result["baseHP"] = stats["BaseHP"] + stats["BaseHPAdd"] * (level - 1)
    result["baseAttack"] = stats["BaseAttack"] + stats["BaseAttackAdd"] * (level - 1)
    result["baseDefence"] = stats["BaseDefence"] + stats["BaseDefenceAdd"] * (level - 1)

    return result


def format_stat_values(values: dict[str, float]) -> dict[str, str]:
    """Format the stat values to a human-readable format.

    Percentage values will be rounded to 1 decimal, while others will be rounded to the nearest integer.

    Args:
        values (dict[str, float]): A dictionary of fight prop ID and value.
    """
    result: dict[str, str] = {}

    for fight_prop, value in values.items():
        if fight_prop in PERCENTAGE_FIGHT_PROPS:
            # round to 1 decimal
            result[fight_prop] = f"{value * 100:.1f}%"
        else:
            result[fight_prop] = str(round(value))

    return result


T = TypeVar("T")


def replace_fight_prop_with_name(
    values: dict[str, T], manual_weapon: dict[str, str]
) -> dict[str, T]:
    """Replace the fight prop with the corresponding name.

    Manual weapon example: https://api.ambr.top/v2/cht/manualWeapon

    Args:
        values (dict[str, T]): A dictionary of fight prop ID and value.
        manual_weapon (dict[str, str]): A dictionary from project ambr with fight prop ID and value.
    """
    return {
        manual_weapon.get(fight_prop, fight_prop): value for fight_prop, value in values.items()
    }


def get_skill_attributes(descriptions: list[str], params: list[int | float]) -> str:
    """Get the skill attributes from the descriptions.

    Args:
        descriptions (list[str]): The list of descriptions.
        params (list[int | float]): The list of parameters.
    """
    result = ""
    for desc in descriptions:
        try:
            k, v = replace_params(desc, params)
        except ValueError:
            continue
        result += f"{k}: {v}\n"
    return result


def remove_ruby_tags(text: str) -> str:
    """Remove ruby tags from a string."""
    # Remove {RUBY_E#} tags
    text = re.sub(r"\{RUBY_E#\}", "", text)
    # Remove {RUBY_B...} tags
    text = re.sub(r"\{RUBY_B#.*?\}", "", text)
    return text
